extern const int ncoef_EGM96;
extern const int n_EGM96[65338];
extern const int m_EGM96[65338];
extern const double c_EGM96[65338];
extern const double s_EGM96[65338];
extern const int ncoef_GRGM360;
extern const int n_GRGM360[65340];
extern const int m_GRGM360[65340];
extern const double c_GRGM360[65340];
extern const double s_GRGM360[65340];
extern const int ncoef_MRO120F;
extern const int n_MRO120F[7380];
extern const int m_MRO120F[7380];
extern const double c_MRO120F[7380];
extern const double s_MRO120F[7380];
